
#define IMPROVE

#define _goodheight_weight  56
#define _gaps_weight        17
#define _edges_weight       170
#define _startpunishgap     4
#define _full_row_weight    20
#define _bottom_well_weight 1

/* alternative
goodheight_weight = 48
gaps_weight = 7
edges_weight = 160
startpunishgap = 8
full_row_weight = 80
bottom_well_weight = 1
*/


#include "game.h"
#include <stdio.h>

static char last_y = WELL_DEPTH;
static char best_x = 0;
static char best_orient = 0;
static WORD new_best_score = 0;
static char try_x = 0;
static BYTE try_left = 0;
static BYTE try_orient = 0;
static BYTE tries = 0;
static BYTE rot_attempt = 0;
static BYTE do_drop = 0;
DWORD arng_value;

#define START_SCORE 32000 
#define NO_SCORE    100

#define THINKING 1
#define DONE 2
#define MOVING 3
#define COVERED 0x10
#define PLACED 0x20

#ifdef TUNING
WORD goodheight_weight = _goodheight_weight;
WORD gaps_weight = _gaps_weight;
WORD edges_weight = _edges_weight;
WORD startpunishgap = _startpunishgap;
WORD bottom_well_weight = _bottom_well_weight;
WORD full_row_weight = _full_row_weight;
#else
#define goodheight_weight _goodheight_weight
#define gaps_weight _gaps_weight
#define edges_weight _edges_weight
#define startpunishgap _startpunishgap
#define bottom_well_weight _bottom_well_weight
#define full_row_weight _full_row_weight
#endif

static BYTE arng(void)
{
    arng_value = ((((DWORD) arng_value & (DWORD) 0xffffL) * (DWORD) 18000L) 
            + ((DWORD) arng_value >> (DWORD) 16L));
    return (BYTE) arng_value;
}

static BYTE try_place(char x, char y, 
        const char fig[SQUARE_NUM][2])
{
    char fx, fy;
    BYTE i;
    
    i = fx = fy = 0;

    for (i = 0; i < SQUARE_NUM; i++) {
        fx = x + fig[i][0];
        fy = y + fig[i][1];
        if (fy < 0) {
            // No collision with the top of the well
            continue;
        }

        if ((fy >= WELL_DEPTH)
        || (fx < 0)
        || (fx >= WELL_WIDTH)) {
            // Collision with the walls of the well
            return 0;
        }
        
        if (get_well(fy, fx) == 1) {
            // Collision with existing block
            return 0;
        }
    }

    // No collision detected
    return 1;
}

static WORD find_edges(char x, char y)
{
    WORD count = 0;

    if ((y >= WELL_DEPTH)
    || (x < 0) || (y < 0)
    || (x >= WELL_WIDTH)) {
        return 0;
    }
    if (y == WELL_DEPTH - 1) {
        count += bottom_well_weight;
    } else {
        if (get_well(y + 1, x) == 1) {
            count += 1;
        }
    }

    if((x == WELL_WIDTH - 1) || (x == 0)) { // if next to wall of well
        count += 1;
    } else {
        if (get_well(y, x + 1) == 1) {
            count += 1;
        }
        if (get_well(y, x - 1) == 1) {
            count += 1;
        }
    }

    return count;
}

static WORD Get_Gaps(char x, char y)
{
    WORD count = 0;
    WORD countdown = startpunishgap;
    char ypos = y+1;
    char yto = ypos + 6;
    char i;

    if ((x < 0) || (x >= WELL_WIDTH)) {
        return 0;
    }
    if (ypos < 0) { 
        ypos = 0; 
    }
    for (i = ypos; i < yto; i++) {
        if((i >= WELL_DEPTH) 
        || (get_well(i, x) == 1)) {
            break;
        } else {
            count = count + countdown; // penalise for leaving empty block below (unless piece takes up
            countdown+= 1; //getting harsher for each row down
        }
    }

    return count;
}

static void think(const char fig[SQUARE_NUM][2])
{
    BYTE cur_block = tet_state.i;
    char cur_y = tet_state.y;
    BYTE max_rot = ROTATIONS[cur_block];
    char y = 0;
    char good_depth = -1;
    WORD score = START_SCORE;

    char deepest = 0;
    char shallowest = WELL_DEPTH - 1;
    WORD edges = 0;
    WORD gaps = 0;
    BYTE full_row = 0;
    BYTE j, k;

    if (cur_y < 0) cur_y = 0;

    for (y = cur_y; y < WELL_DEPTH; y++) {
        // replace with int[10] every move for top 10 parts of game area, then use this
        if (try_place(try_x, y, fig)) {
            good_depth = y;
        } else {
            break;
        }
    }

#ifdef DEBUG
    printf("n.b. try remaining %u good depth %u try_x %u r %u\n",
                tries, good_depth, try_x, try_orient);
#endif

    if (good_depth > 0) {

        // find out how many edges the piece would be adjacent to?
        for (j = 0; j < SQUARE_NUM; j++) {
            char xp = fig[j][0];
            char yp = fig[j][1];
            char x_pos_in_game, y_pos_in_game;
            char below = 1;

            x_pos_in_game = xp + try_x;
            y_pos_in_game = yp + good_depth;

            // check if any of the piece is directly below this piece
            // always use bottom-most part
            while (1) {
                BYTE blockbelow = 0;
                for (k = 0; k < SQUARE_NUM; k++) {
                    if ((xp == fig[k][0]) 
                    && ((yp+below) == fig[k][1])) {
                        blockbelow = 1;
                    }
                }
                if (blockbelow) {
                    below++;
                } else {
                    break;
                }
            } 
            gaps += Get_Gaps(x_pos_in_game, y_pos_in_game + below - 1);

            edges += find_edges(x_pos_in_game, y_pos_in_game);

            if (y_pos_in_game > deepest) {
                deepest = y_pos_in_game;
            }
            if ((y_pos_in_game > 0) && (y_pos_in_game < shallowest)) {
                shallowest = y_pos_in_game;
            }
        }
        // Check for rows that will become complete
        // We aim for these as a priority
#ifdef IMPROVE
        for (y = shallowest; y <= deepest; y++) {
            BYTE row_complete = 1;
            BYTE x;
            for (x = 0; x < WELL_WIDTH; x++) {
                BYTE filled = 0;
                if (get_well(y, x) == 1) {
                    filled = 1;
                } else {
                    for (k = 0; k < SQUARE_NUM; k++) {
                        if ((fig[k][1] == x)
                        && (fig[k][0] == y)) {
                            filled = 1;
                        }
                    }
                }
                if (!filled) {
                    row_complete = 0;
                }
            }
            if (row_complete) {
                full_row ++;
            }
        }
#endif
        
        score += (edges*edges_weight) + ((WORD)deepest*goodheight_weight) +
                    ((WORD)full_row*full_row_weight) - (gaps*gaps_weight);
#ifdef DEBUG
        printf("n.b. score %u with x %u o %u edges %u "
                    "deepest %u full_row %u gaps %u (%u, %u)\n", 
                score, try_x, try_orient,
                (edges*edges_weight), ((WORD)deepest*goodheight_weight),
                ((WORD) full_row*full_row_weight),
                (gaps*gaps_weight), try_x, good_depth);
#endif
    } else {
        // some part was outside of well set to less than best score so far
        score = new_best_score - 2;
#ifdef DEBUG
        printf("n.b. outside with x %d o %u\n",
                try_x, try_orient);
#endif
    }
    if (score > new_best_score) {
        new_best_score = score;
        best_orient = try_orient;
        best_x = try_x;
#ifdef DEBUG
        printf("n.b. new best %u with x %u o %u\n", 
                score, best_x, best_orient);

#endif
    }

    if (try_orient >= max_rot) {
        if (try_left) {
            if (try_x == 0) {
                try_x = WELL_WIDTH - 1;
            } else {
                try_x --;
            }
        } else {
            try_x ++;
            if (try_x >= WELL_WIDTH) {
                try_x = 0;
            }
        }
        try_orient = 0;
        tries --;
    } else {
        try_orient ++;
    }
}

Player_Move ai_move(void)
{
    BYTE cur_block = tet_state.i;
    BYTE cur_orient = tet_state.r;
    char cur_x = tet_state.x;
    char cur_y = tet_state.y;
    BYTE max_rot = ROTATIONS[cur_block];
    Player_Move move = MOVE_NONE;
    BYTE hard = (player_moves_per_down_move <= 2);
    BYTE i;

    if (cur_y < last_y) {
        // New block
        new_best_score = NO_SCORE;
        i = arng();
        try_x = i % WELL_WIDTH;
        try_left = i & 0x80;
        try_orient = 0;
        tries = WELL_WIDTH;
        rot_attempt = 0;
        do_drop = hard ? 1 : 5;
#ifdef DEBUG
        printf("n.b. AI start new block i %u r %u\n",
                tet_state.i, tet_state.r);
#endif

        best_x = cur_x;
        best_orient = cur_orient;
        think(TETRITABLE[(BYTE)cur_block][(BYTE)try_orient]);
        think(TETRITABLE[(BYTE)cur_block][(BYTE)try_orient]);
    } else if (tries) {
        i = 3;
        while(i && tries) {
            think(TETRITABLE[(BYTE)cur_block][(BYTE)try_orient]);
            i--;
        }
    } else {
        i = arng();
    }

    last_y = cur_y;
    if (cur_y <= 1) {
        // No moves until the piece is on screen (I mean it!)
        return MOVE_NONE;
    }

    // Move the block into the position decided by the THINKING phase.
    // Try to rotate first
    if ((rot_attempt <= max_rot)
    && ((cur_orient & max_rot) != (best_orient & max_rot))) {
        move = MOVE_ROTATE;
        rot_attempt ++;
    } else {
        // Already rotated? Then move
        if (cur_x < best_x) {
            move = MOVE_RIGHT;
        } else if (cur_x > best_x) {
            move = MOVE_LEFT;
        } else if ((rot_attempt == 0) && (tries == 0)) {
            if (do_drop == 0) {
                move = MOVE_DROP;
            } else {
                do_drop--;
            }
        }
        rot_attempt = 0;
    }
#ifdef DEBUG
    printf("n.b. AI move %u\n", move);
#endif

    if ((!hard) 
    && (((cur_y < 6) && (i & 0x20))
    || ((cur_y < 9) && (i & 0x40))
    || ((cur_y < 12) && (i & 0x80)))) {
        move = MOVE_NONE;
    }

    if (move == MOVE_DROP) {
        last_y = WELL_DEPTH;
    }

    return move;
}

